#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "bellmanford.h"

//#define PRINT_CYCLE

//   ORIGINAL VERSION   //
/*   
void 
bellmanford(unsigned int numnode, graph_t *E, float *w)
{
  unsigned int i;
  edge_t *edge;

  w[0] = 0;
  for (i=1; i<=numnode; i++) 
    w[i] = P_INFINITY;
  for (i=1; i<=numnode; i++) {
    edge = E->head;
    while (edge) {
      if ((w[edge->sn] < P_INFINITY/2 ) && (w[edge->en]-w[edge->sn]-((info_t *)edge->info)->w > EPSI)) {
      	w[edge->en] = w[edge->sn] + ((info_t *)edge->info)->w;
      }
      edge = edge->next;
    }
  }
}
*/   

//   IMPROVED VERSION    //
unsigned int
bellmanford(unsigned int numnode, edge_t **Adj, float *w)
{
  unsigned int i, u, v, u1, *p, *rank, firstelem, fail=0;
  float *pi;
  list_t *list = list_create();
  list_t *list1 = list_create();
  list_t *tmp;
  edge_t *edge;
	  
  if ((pi = (float *)calloc(numnode+1, sizeof(float))) == NULL) no_mem_exit();
  if ((p = (unsigned int *)calloc(numnode+1, sizeof(unsigned int))) == NULL) no_mem_exit();
  if ((rank = (unsigned int *)calloc(numnode+1, sizeof(unsigned int))) == NULL) no_mem_exit();

  for (i=0; i<=numnode; i++)
    w[i] = 0;
  list_append(list, (void *)0);
  do {
    for (i=0; i<=numnode; i++)
      set_make(p, rank, i);
    firstelem = numnode+1;
    while ((list_length(list) > 0) && !fail) {
      u = (unsigned int)list_extracthead(list);	
      if (u == 0) {    // 0 is a dummy.
        for (i=1; i<=numnode; i++) {
          list_append(list1, (void *)i);
          pi[i] = 0;
        }
      } 
      else {           // non-dummy vertex.
        edge = Adj[u];
        while (edge) {
	  v = edge->en;
          if (w[v]-w[u]-((info_t *)edge->info)->w >EPSI) {
	    w[v] = w[u] + ((info_t *)edge->info)->w;      
            pi[v] = u;
            if (list_length(list1) == 0) {
              list_append(list1, (void *)v);
              firstelem = v;
            } 
	    else if (set_find(p, firstelem) 
		     != set_find(p, v)) {
              list_append(list1, (void *)v);
              set_union(p, rank, firstelem, v);               
            }
	  }
	  edge = edge->nextadj;
        }
        if ((firstelem != numnode+1) && 
	    (set_find(p, firstelem) == set_find(p, 0))) 
          fail = 1;
        if (fail == 0) {
          u1 = u;
          while ((pi[u]!=u1) && (pi[u]!=0)) 
            u = pi[u];
          if (pi[u] == u1) {
            fail = 1;
#ifdef PRINT_CYCLE
	    u1 = u;
	    printf("%d <----- ", u);
	    while ((pi[u]!=u1) && (pi[u]!=0)) {
	      u = pi[u];
	      printf("%d <----- ", u);		 
	    }
	    printf("%d\n", u1);
#endif
	  }
        }
      }      
    }
  
    if (!fail && (list_length(list1) > 0)) {
      tmp = list;
      list = list1;
      list1 = tmp;
    }
  } while ((list_length(list) > 0) && !fail);
  
  list_free(list);
  list_free(list1);
  free(p);
  free(rank);
  free(pi);
  return fail;
}



