#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "retiming.h"

void retiming_restore(graph_t *E, unsigned int *pbak)
{
  edge_t *edge;
  unsigned int i = 0;

  edge = E->head;
  while (edge) {
    ((info_t *)edge->info)->p = pbak[i++];
    edge = edge->next;
  }
}

void delete_extra_edge(graph_t *E, graph_t *hE, 
		       unsigned int numnode)
{
  edge_t *edge, *tmp=0;
  edge = E->head;
  E->edgeno = 0;
  while ((edge->sn!=numnode+1) && (edge->en!=numnode+1)) {
    tmp = edge;
    edge = edge->next;
    E->edgeno++;
  }
  if (tmp)
    tmp->next = 0;
  hE->head = edge;
}

unsigned int
retiming_compute(unsigned int numnode, graph_t *E, graph_t *hE, 
		 fibheap_t *fh, float lamda, edge_t **Adj, 
		 unsigned int *rEedgeNo)
{
  unsigned int num, i, j, fail = 0, numedge;
  float *w, *dij, *wbak, judge;
  edge_t *edge, *tmp;
  graph_t *rE = graph_create();
  info_t *info;

  num = numnode + 1;
  numedge = E->edgeno;
  if ((w = (float *)calloc(num, sizeof(float))) == NULL) no_mem_exit();
  if ((wbak = (float *)calloc(numedge, sizeof(float))) == NULL) no_mem_exit();
  if ((dij = (float *)calloc(num, sizeof(float))) == NULL) no_mem_exit();

  i = 0;
  edge = E->head;
  while (edge) {
    wbak[i++] = ((info_t *)edge->info)->w;
    ((info_t *)edge->info)->w = ((info_t *)edge->info)->w 
      - ((info_t *)edge->info)->t*lamda;
    ((info_t *)edge->info)->w = -1 * ((info_t *)edge->info)->w;
    edge = edge->next;
  }  

  fail=bellmanford(numnode, Adj, w);
  if (fail) {
    printf("\nPositive Cycle for sd(u,v).\n");
    mcrp_restore(E, wbak);
    graph_free(rE);
    free(w);
    free(wbak);
    free(dij);
    return 0;
  }

  edge = E->head;
  while (edge){
    ((info_t *)edge->info)->w = ((info_t *)edge->info)->w + w[edge->sn] - w[edge->en];
    edge = edge->next;
  } 
  // All-pair longest path computation.
  for (i=1; i<=numnode; i++) {
    dijkstra(numnode, E, fh, dij, i, Adj);
    for (j=1; j<=numnode; j++) 
      if ((dij[j] < P_INFINITY / 2) && (i != j)) {
	dij[j] = -1 * (dij[j] - w[i] + w[j]);
	dij[j] = ceil(dij[j]/lamda) - 1;
	
	if ((info = (struct info *)calloc(1, sizeof(struct info))) == NULL) no_mem_exit();
	info->w = -dij[j];
	info->t = 1;  // t=1 for Howard
	info->p = 0;
	edge = graph_add_edge(rE, j, i, (void *)info);
      }
  }
  edge = E->head;
  while (edge) {
    if (((info_t *)edge->info)->p == 0) {
      i = edge->sn;
      j = edge->en;
      if ((info = (struct info *)calloc(1, sizeof(struct info))) == NULL) no_mem_exit();
      info->w = 0;
      info->t = 1;  // t=1 for Howard
      info->p = 0;
      tmp = graph_add_edge(rE, i, j, (void *)info);
      
      if ((info = (struct info *)calloc(1, sizeof(struct info))) == NULL) no_mem_exit();
      info->w = 0;
      info->t = 1;  // t=1 for Howard
      info->p = 0;
      tmp = graph_add_edge(rE, j, i, (void *)info);
    }
    edge = edge->next;
  }
  // Howard's mcrp algorithm to test feasibility.
  edge = hE->head;
  while (edge) {
    i = edge->sn;
    j = edge->en;
    if ((info = (struct info *)calloc(1, sizeof(struct info))) == NULL) no_mem_exit();
    info->w = 0;
    info->t = 1;  // t=1 for Howard
    info->p = 0;
    tmp = graph_add_edge(rE, j, i, (void *)info);
    if ((info = (struct info *)calloc(1, sizeof(struct info))) == NULL) no_mem_exit();
    info->w = 0;
    info->t = 1;  // t=1 for Howard
    info->p = 0;
    tmp = graph_add_edge(rE, i, j, (void *)info);
    edge = edge->next;
  }	  
  mcrp_compute(numnode + 1, rE, &judge, 1);
  if (judge < 0)
    fail = 1;
  
  *rEedgeNo = rE->edgeno;
  graph_free(rE);
  mcrp_restore(E, wbak);
  free(w);
  free(wbak);
  free(dij);
  if (fail)
    return 0;
  return 1;
}


