#include <stdio.h>
#include <stdlib.h>
#include "retiming.h"

#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>

#ifndef HZ
#define HZ     CLOCKS_PER_SEC
#endif

void get_cpu_time(double *usertime, double *systime) {
        struct tms time;
        times( &time );
        *usertime = (double)time.tms_utime / (double)HZ;
        *systime  = (double)time.tms_stime / (double)HZ;
}

static float 
compare(void *data1, void *data2)
{
  float tmp;
  tmp = *(((data_t *)data1)->v) - *(((data_t *)data2)->v);
  return tmp;
} 
    
static void 
free_fcn(void *data)
{   
  free(data); 
}   

int main()
{
  // Variable declaration (elaborated when used).
  unsigned int numnode, numedge, numpi, numpo, pi, po, u, v, 
    t, p, i, rEedgeNo, num, *discover, *finish, *pbak;
  int ok, times = 0, *optr, maxdif;
  float w, maxgate = 0, lamda, upper, neg = NEGINFINITY, 
    *delta, maxblock=0;
  graph_t *E = graph_create();
  graph_t *hE = graph_create();
  graph_t *E0 = graph_create();
  list_t *qlist =  list_create();
  list_t *topolist = list_create();
  list_t **inlist;
  list_item *item=0;
  edge_t *edge, **Adj, **E0Adj, **revE0Adj, *tmp;
  info_t *info;

  data_t *data;
  data = (data_t *)calloc(1, sizeof(data_t));
  data->v = &neg;
  data->id = 0;
  fibheap_t *fh = fibheap_create(compare, data);
//  FILE *fp;
//  fp=fopen("/home/clin/src/howardretiming/testfile","rb");

  scanf("%d%d%d%d", &numnode, &numedge, &numpi, &numpo);
  num = numnode + 1;
  if ((Adj = (edge_t **)calloc(num, sizeof(edge_t *))) == NULL) no_mem_exit();
  if ((E0Adj = (edge_t **)malloc(num*sizeof(edge_t *))) == NULL) no_mem_exit();
  if ((revE0Adj = (edge_t **)malloc(num*sizeof(edge_t *))) == NULL) no_mem_exit();
  if ((delta = (float *)calloc(num, sizeof(float))) == NULL) no_mem_exit();
  if ((discover = (unsigned int *)calloc(num, sizeof(unsigned int))) == NULL) no_mem_exit();
  if ((finish = (unsigned int *)calloc(num, sizeof(unsigned int))) == NULL) no_mem_exit();
  if ((optr = (int*)calloc(num, sizeof(int))) == NULL) no_mem_exit();
  if ((inlist = (list_t **)malloc(num*sizeof(list_t *))) == NULL) no_mem_exit();
  for (i=1; i<=numnode; i++) {
    Adj[i] = 0;
    E0Adj[i] = 0;
    revE0Adj[i] = 0;
    delta[i] = 0;
    discover[i] = 0;
    finish[i] = 0;		
    inlist[i] = list_create();
  }

  // Circuit read-in.
  for (i=0; i<numpi; i++) {
    scanf("%d", &pi);
    if ((info = (struct info *)calloc(1, sizeof(struct info))) == NULL) no_mem_exit();
    info->w = 0;
    info->t = 1;
    info->p = 1;
    edge = graph_add_edge(E, numnode+1, pi, (void *)info);
  }
  for (i=0; i<numpo; i++) {
    scanf("%d", &po);
    if ((info = (struct info *)calloc(1, sizeof(struct info))) == NULL) no_mem_exit();
    info->t = 0;
    info->p = 1;
    info->w = 0;
    edge = graph_add_edge(E, po, numnode+1, (void *)info);
  }
  for (i=0; i<numedge; i++) {
    scanf("%d%d%f%d%d", &u, &v, &w, &t, &p);
    if ((p == 0) && (w > maxgate)) 
      maxgate = w;
    if ((info = (struct info *)calloc(1, sizeof(struct info))) == NULL) no_mem_exit();
    info->t = t;
    info->p = p;
    info->w = w;
    edge = graph_add_edge(E, u, v, (void *)info);
    graph_adjacent(Adj, edge);
  }

  // Compute lower bound.
  mcrp_compute(numnode+1, E, &lamda, 0); 
  printf("\nLower Bound Period: %f\n", lamda);
 
  // Compute upper bound: 
  // T_u = T_l + maxblockdelay +  T_l*FFdiscrepancy
  // Step1: Use topological sort to compute the maxblockdelay.
  // During the process construct dependency list of each node.  
  upper = lamda;
  edge = E->head;
  while (edge) {
    if ((((info_t *)edge->info)->p == 0) 
	&& (((info_t *)edge->info)->t == 0)) {
      if ((info = (struct info *)calloc(1, sizeof(struct info))) == NULL) no_mem_exit();
      info->w = ((info_t *)edge->info)->w;
      info->t = ((info_t *)edge->info)->t;
      info->p = ((info_t *)edge->info)->p;
      // E0 is the subgraph of E where edges have no FFs.
      tmp = graph_add_edge(E0, edge->sn, edge->en, (void *)info); 
      graph_adjacent(E0Adj, tmp);
      graph_adjacent_en(revE0Adj, tmp);
    } else if ((((info_t *)edge->info)->p == 0)
	       && (((info_t *)edge->info)->t > 0)) {
      // qlist records the edges transformed from in-block FFs.
      list_append(qlist, (void *)edge->sn); 
      list_append(qlist, (void *)edge->en);
    }
    edge = edge->next;
  }

  // Get the topo order of the vertices in E0.
  dfs(numnode, E0, E0Adj, discover, finish, topolist); 
  free(discover);
  free(finish);
/*
  item = topolist->head;
  while (item) {
    printf("%d ", (int)item->data);
    item = item->next;
  }
  printf("\n");
*/
  while (list_length(topolist) > 0) {
    v = (unsigned int)list_extracthead(topolist);
    if ((revE0Adj[v] == 0) && (E0Adj[v] != 0)) {
      //initialize the nodes with no incoming edges.
      list_append(inlist[v], (void *)v);  
    } else if (revE0Adj[v] != 0) {
      edge = revE0Adj[v];
      while (edge) {
	if (delta[edge->sn] + ((info_t *)edge->info)->w > delta[v]) 
	  //delta[v] is the arrival time of v
	  delta[v] = delta[edge->sn] + ((info_t *)edge->info)->w;
	item = inlist[edge->sn]->head;
	while (item) {
	  u = (unsigned int)item->data;
	  // inlist[v] records inputs of the block that v is in.
	  list_append(inlist[v], (void *)u);  
	  item = item->next;
	}
	edge = edge->prevadj;
      }
    }
  }
  for (i=1;i<=numnode;i++) {
    if (maxblock < delta[i])
      maxblock = delta[i];
    delta[i] = 0;
  }
  printf("max block delay: %f\n",maxblock);
  upper = upper + maxblock;	  

/* 
  // Update the dependency list due to edges transferred from in-block FFs. //
  while (list_length(qlist) > 0) {
    u = (unsigned int)list_extracthead(qlist);
    v = (unsigned int)list_extracthead(qlist);
    item = inlist[u]->head;
    while (item) {
      i = (unsigned int)item->data;
      list_append(inlist[v], (void *)i);
      item = item->next;
    }
    edge = E0Adj[v];
    while (edge) {
      list_append(qlist, (void *)edge->sn);
      list_append(qlist, (void *)edge->en);
      edge = edge->nextadj;
    }
  }
  for (i=1;i<=numnode;i++) {
    if (E0Adj[i] != 0)
      list_freeitems(inlist[i]);     
  }
*/
  // Step2: Find the r value under T_l w/o FF restriction. 
  // After that, find the max FF discrepancy of adjacent edges.
  // restore E.
  // hE contains the host vertex M and its incident edges.
  delete_extra_edge(E, hE, numnode);  
 
  if ((pbak = (unsigned int *)calloc(E->edgeno, sizeof(unsigned int))) == NULL) no_mem_exit();
  i = 0;
  edge = E->head;
  while (edge) {
    pbak[i++] = ((info_t *)edge->info)->p;
    // Make the edge flip-flop allowable.
    ((info_t *)edge->info)->p = 1;   
    edge = edge->next;
  }
  // Compute the lower bound T2.
  ok = retiming_compute(numnode, E, hE, fh, lamda*1.01, 
			Adj, &rEedgeNo);  
  if (ok == 0) { 
    printf("A valid retiming should exist without forbidden edges.\n");
    exit(0);
  }
  retiming_restore(E, pbak);
  free(pbak);

  maxdif = 0;
  for (i=1; i<=numnode; i++) {
    if (list_length(inlist[i]) > 0) {
      while (list_length(inlist[i]) > 0) {
        u = (unsigned int)list_extracthead(inlist[i]);
	if (optr[i]-optr[u] > maxdif)
          maxdif = optr[i]-optr[u];
      }
    }
  }
  if (maxdif < 0) { 
    printf("FF discrepancy should be >= 0\n"); 
    exit(0);
  }
  upper = upper + maxdif*lamda; 
  printf("max r difference: %d\n",maxdif);
  printf("Upper Bound Period: %f\n\n", upper);
  
  list_free(qlist);
  list_free(topolist);
  for (i=1;i<=numnode;i++)
    list_free(inlist[i]);
  free(inlist);
  graph_free(E0);
  free(E0Adj);
  free(revE0Adj);
  free(delta);
  free(item);
    
  // Process binary search.
  double utime, stime, utime1, stime1;
  get_cpu_time(&utime, &stime);

  do { 
    ok = retiming_compute(numnode, E, hE, fh, (upper+lamda)/2,
			  Adj, &rEedgeNo);
    times ++;
    printf("binary_upper=%f, binary_lower=%f, midpoint feasible?(1/0): %d\n",upper,lamda,ok);
    printf("rE Constraints No: %d\n", rEedgeNo);
    if (ok) 
      upper = (upper+lamda)/2;     
    else
      lamda = (upper+lamda)/2;
  } while (upper-lamda > PRECISION);  

  // Output results
  get_cpu_time(&utime1, &stime1);
  printf("\nOptimal Period: %f\nPrecision: %f\n", upper, PRECISION);

//  for (i=1; i<=numnode; i++)
//    printf("r[%d]=%d\t",i, (int)(optr[i] - optr[numnode + 1]));
//  printf("\n"); 

  printf("No. of iterations: %d\n",times);
  printf("Total CPU time:\t%4.3f u  %4.3f s\t%4.3f total (sec)\n",
              utime1-utime, stime1-stime, utime1 + stime1-utime-stime);
      
  fibheap_free(fh, free_fcn);
  graph_free(E);
  graph_free(hE);
  free(Adj);
  free(optr);
  return 0;
}  
